package edu.harvard.mgh.purcell.gplink.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import java.util.HashMap;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import org.apache.log4j.Logger;

import edu.harvard.mgh.purcell.gplink.mainGUI.*;

/**
 * AutoUpadater extends Timer, automatically updating
 * the output files attached in OpView.
 * @author Kathe Todd-Brown
 *
 */
public class AutoUpdater extends Timer {
	
	/**
	 * Define a static logger variable
	 */
	private static Logger logger = Logger.getLogger(AutoUpdater.class);

	/**
	 * A string that flags a completed operation
	 */
	public static String COMPLETE = "Complete";
	/**
	 * A string that flags a running operation
	 */
	public static String RUNNING = "Running";
	/**
	 * A string that flags a failed operation
	 */
	public static String ERROR = "Failed";
	
	/**
	 * A hash map that keeps all the operations and their current status
	 * completed, running, or failed.
	 */
	public HashMap<String, String> opStatis;
	
	/**
	 * The time between schedule runs
	 */
	private long interval;
	
	/**
	 * This classes access to the general data
	 */
	private Project data;
	
	/**
	 * This runs every N seconds ideal but can be delayed if there isn't
	 * enough processor space.
	 * If the operation is still running then check for a .gplink file
	 * for the completion status. If a running operation is now completed
	 * get the current list of files in our project folder and compares
	 * them to the files listed as operations outputs. If there are files
	 * that are not in the output list but matches the operation root then
	 * it is added to the correct operation output.
	 */
	private TimerTask readLogs = new TimerTask(){
		@SuppressWarnings("static-access")
		public void run() {
			//tell the logger what's happening
			logger.info("Entering readLogs()");
			//Error out if the operation status is null. This
			//...should never happen.
			if(opStatis == null){
				//if we are null then tell the user that there's a problem
				new ErrorDialog(data, "Hash table of operation status" +
						" is null, discontinuing" +
						" auto update feature.\n run():AutoUpdater.java");
				//tell the logger there's a problem
				logger.error("opstatis is null. Canceling auto updater.");
				//cancel the auto updater
				AutoUpdater.this.cancel();
				//leave the function
				return;
			}
			//get a list of all the files in our project
			Vector <String> allFiles = data.getFileList();
			//get a list of all the operations in the tree
			Vector <String> allOps = data.getAllOpName();
			//check that the operations and files vectors 
			//...are all non-null 
			if(allOps == null || allFiles == null){
				//tell the user something is wrong
				new ErrorDialog(data, "Error! Can not access the operations " +
						"or files correctly." +
						" Stopping auto updating feature.\n " +
						"run()@AutoUpdater.java");
				//log it
				logger.error("The operations or file vectors are null!" +
						" Turning off auto updater.");
				//cancel the auto updater
				cancel();
				//leave the function
				return;
			}
			
			//
			//////Download the new log and gplink from the remote directory
			//
			
			if(data.isRemote()){
				//keep track of the files we need to download
				Vector <String> filesToDownload = new Vector<String>();
				//go through all the operations
				for(String s: allOps){
					//if the operation isn't null then get the "status"
					//...further more if hte status is either null (ie not in
					//...the hash table or "RUNNING" then look for files belonging
					//...to that operaiton
					if(s != null
							&& (opStatis.get(s) != null
							|| opStatis.get(s).equals(RUNNING))){
						//check the files for _opname.gplink_ files
						if(allFiles.contains(
							data.getRemoteFile() + "/" + s + data.OP_STATUS_EXT)){
							//yea! we found one... flag it for download
							filesToDownload.add(
								data.getRemoteFile() + "/" + s + data.OP_STATUS_EXT);
						}
						//check the files for _opname.log_ files
						if(allFiles.contains(
								data.getRemoteFile() + "/" + s + data.OP_LOG_EXT)){
							//yea! we found one! flag it for download
							filesToDownload.add(
								data.getRemoteFile() + "/" + s + data.OP_LOG_EXT);
						}
					}
				}//end for loop looking at operations
				
				//get the number of files we need to download
				int numdownloads = filesToDownload.size();
				//if there are any files to download
				if(numdownloads > 0){
					//move the vector into an array
					String [] downloadFile = new String [numdownloads];
					for(int i = 0; i < numdownloads; i ++){
						downloadFile[i] = filesToDownload.get(i);
					}
					//download the files using our download thread
					DownloadFile downloader = 
						new DownloadFile(downloadFile, 
								data.getRoot().getAbsolutePath(), 
								data.getConn(), data.frame);
					//start the thread
					downloader.start();
					try { 
						//wait for the download to finish so we can see the
						//...exit status for the operations
						downloader.join();
					} catch (InterruptedException e) {
						//tell the user there is an error
						new ErrorDialog(data.frame, "Error trying to" +
								" download files.\n run().readLogs@AutoUpdater.java");
						//tell the logger something happened
						logger.error("Interuption of log/gplink download");
					}
				}//end if-download
			}//end if-remote
			
			//flag to see if anything has changed in the tree -> need to
			//...update the main frame viewers
			boolean change = false;
			
			//Go through each operation and check the status
			for(String s: data.getAllOpName()){
				//Check that the operation is valid and is marked as running
				if(s != null
						&&( opStatis.get(s) == null 
								|| opStatis.get(s).equals(RUNNING))){
				
					//the file that contanis the status of the operation
					File gplinkLog = new File(data.getRoot(), s + data.OP_STATUS_EXT);
					
					//if this file exist then the command/operation is done
					if(gplinkLog.exists()){
						//initalize the setting to fail (if there was some error writing
						//...the gplink log then we have a default failure
						String setting = "1";
						//read in the content of the file
						BufferedReader in;
						try {
							in = new BufferedReader(new FileReader(gplinkLog));
							setting = in.readLine();
							in.close();
						} catch (IOException e) {
							new ErrorDialog(data.frame, "Error trying to read " +
									" [ " + gplinkLog + " ], " +
									"["+ s +"] being flagged as unsucessful. \n " +
									"readLogs:AutoUpdater.java");
							logger.error("Trouble reading [" + gplinkLog +"].");
						}
						
						//0 is a sucessful completion
						if(setting.equals("0")){
							opStatis.put(s, COMPLETE);
						} else {
							opStatis.put(s, ERROR);
						}
						//we changed the tree!
						change = true;
					} else{
						//the operation is still running
						opStatis.put(s, RUNNING);
					}
				}//end if-RUNNING check
			}//end for all operaitons
			
			//did we change something in the hash table?
			if(change == true){
				//update the files associated with the project 
				//...and our displays
				data.frame.updateViews();
			}
			
			//auto-saving feature
			data.writeProjectInfo(true, false, false);
			//tell the logger we are done
			logger.info("Exiting readLogs()");
		}
		
	};
	
	/**
	 * Add an operation to our updater and flag it's status correctly.
	 * @param s a String naming the operation
	 */
	@SuppressWarnings("static-access")
	public void addOp(String s){
		
		File gplinkLog = new File(data.getRoot(), s + data.OP_STATUS_EXT);
		//if the log exists then read in the status and flag it
		if(gplinkLog.exists()){
			BufferedReader in;
			try {
				in = new BufferedReader(new FileReader(gplinkLog));
				String setting = in.readLine();
				in.close();
				if(setting.contains("0"))
					opStatis.put(s, COMPLETE);
				else
					opStatis.put(s, ERROR);
				
			} catch (IOException e) { 
				//you must have a currupted file
				new ErrorDialog(data.frame, "Error trying to read " +
						" [ " + gplinkLog + " ], " +
						"["+ s +"] being flagged as unsucessful. \n " +
						"readLogs:AutoUpdater.java");
				logger.error("Trouble reading [" + gplinkLog +"].");
				opStatis.put(s, ERROR);
			}
		}
		else
			//flag it as running
			opStatis.put(s, RUNNING);
		
		//update our views
		//data.frame.updateViews();
	}
	
	/**
	 * runs the Timer's cancel and set the opStatis to null for
	 * garbage collection.
	 */
	public void cancel(){
		opStatis = null;
		super.cancel();
	}
	
	/**
	 * Create and start our unique timer. This will update the
	 * status of our operations on the operations viewer.
	 * @param d the current project
	 */
	@SuppressWarnings("static-access")
	public AutoUpdater(Project d) {
		super();
		
		//tell the user we are starting
		logger.info("Creating AutoUpdater");
		
		///////////////////////////////////
		//*********Initalize***************
		///////////////////////////////////
		
		data = d;

		opStatis = new HashMap<String, String>();
		
		//go through all the operations in the project
		for (String s : data.getAllOpName()) {
			addOp(s);
		}
		
		///////////////////////////////////////
		//***********set timer*****************
		////////////////////////////////////////
		
		String timing = data.getUpdaterIncrament();
		
		if(timing.equals(""))
			return;

		interval = new Integer(timing) * 1000;
		
		//run readLogs every N seconds but you can bump it since
		//...it's not critical if the processor is busy
		schedule(readLogs, interval, interval);
		
		logger.info("Done creating AutoUpdater");
	}

}
