package edu.harvard.mgh.purcell.gplink.data;

import java.io.IOException;
import java.util.Vector;

import org.apache.log4j.Logger;

import com.sshtools.j2ssh.ScpClient;
import com.sshtools.j2ssh.SshClient;

import edu.harvard.mgh.purcell.gplink.mainGUI.ErrorDialog;
import edu.harvard.mgh.purcell.gplink.mainGUI.MainFrame;

/**
 * A thread that downloads files from a remote server via
 * scp.
 * @author Kathe Todd-Brown
 *
 */
public class DownloadFile extends Thread {
	/**
	 * Define a static logger variable so that it references the
	 */
	private static Logger logger = Logger.getLogger(DownloadFile.class);
	
	/**
	 * An array of remote files you wish to download.
	 */
	private String [] remoteFile;
	
	/**
	 * The local target file for ssh.
	 */
	private String localFolder;
	
	/**
	 * An autherized ssh connection.
	 */
	private SshClient conn;
	
	/**
	 * A boolean flag that tells us weather to ignore 
	 * a download error. For example if we aren't sure
	 * if the remote file exists then we don't really care
	 * if the scp program can't find the file to download.
	 */
	private boolean ignore;
	
	/**
	 * Frame tha this download is attached to.
	 */
	private MainFrame frame;
	
	/**
	 * Construct a file download
	 * @param getMe A String array that holds the full names
	 * of the remote files to be downloaded
	 * @param putHere A String that holds the full name of
	 * the local directory to download to.
	 * @param c an authenticated connection
	 * @param f a MainFrame this download is attached to
	 */
	public DownloadFile(String [] getMe, String putHere, 
			SshClient c, MainFrame f){
		super();
		remoteFile = getMe;
		localFolder = putHere;
		conn = c;
		ignore = false;
		frame = f;
	}
	
	/**
	 * Construct a file download
	 * @param getMe A String array that holds the full names
	 * of the remote files to be downloaded
	 * @param putHere A String that holds the full name of
	 * the local directory to download to.
	 * @param c an authenticated connection
	 * @param f a MainFrame this download is attached to
	 * @param b a boolean flag that tells us weather we really
	 * care if the download fails.
	 */
	public DownloadFile(String [] getMe, String putHere, 
			SshClient c, MainFrame f, boolean b){
		super();
		remoteFile = getMe;
		localFolder = putHere;
		conn = c;
		ignore = b;
		frame = f;
	}
	
	/**
	 * Construct a file download
	 * @param getMe A String that holds the full name
	 * of the remote file to be downloaded
	 * @param putHere A String that holds the full name of
	 * the local directory to download to.
	 * @param c an authenticated connection
	 * @param f a MainFrame this download is attached to
	 * @param b a boolean flag that tells us weather we really
	 * care if the download fails.
	 */
	public DownloadFile(String getMe, String putHere,
			SshClient c, MainFrame f, boolean i){
		super();
		remoteFile = new String [1];
		remoteFile[0] = getMe;
		localFolder = putHere;
		conn = c;
		ignore = i;
		frame = f;
	}
	
	/**
	 * Construct a file download
	 * @param getMe A String Vector that holds the full names
	 * of the remote files to be downloaded
	 * @param putHere A String that holds the full name of
	 * the local directory to download to.
	 * @param c an authenticated connection
	 * @param f a MainFrame this download is attached to
	 * @param b a boolean flag that tells us weather we really
	 * care if the download fails.
	 */
	public DownloadFile(Vector<String> filesToDownload, String puthere, SshClient c, MainFrame f, boolean b) {
		int max = filesToDownload.size();
		String [] fileArray = new String[max];
		for(int i = 0; i < max; i ++ ){
			fileArray[i] = filesToDownload.get(i);
		}
		
		remoteFile = fileArray;
		localFolder = puthere;
		conn = c;
		ignore = b;
		frame = f;
	}

	/**
	 * The actual download when this thread is run.
	 */
	public void run(){
		//tell the logger what's going on
		logger.info("Starting downloading");
		//declare the scp client
		ScpClient scp = null;
		try {
			//try to establish a scp client
			scp = conn.openScpClient();
			//oh no's! We ran into an error trying to open the 
			//...connection! 
		} catch (IOException e) {
			//log the error
			logger.error( "Error trying open connection to download file.");
			//let this notify the user that the connection was broke
			frame.data.connectionLost();
			//go out of the download
			return;
		}
		//notify the user we are transfuring files
		frame.changeTransitFile(remoteFile, true);
		//translate the remotefiles from an Array to a string
		String remoteFiles ="";
		for(String file:remoteFile){
			remoteFiles = " ["+file+"]";
		}
		//logs the start of the download
		logger.info("Download "+remoteFiles+" to ["+localFolder+"].");
		//copy the files
		try {
			scp.get(localFolder, remoteFile, false);
		//oh no's again! Something didn't work
		} catch (IOException e1) {
			//if we ignore this failure then we probably aren't
			//...sure that the remote file exists
			if(ignore == false){
				//tell the user that we didn't download
				new ErrorDialog(frame, "Error trying download "+remoteFiles+" to ["+localFolder+"].\n" +
					"run@DownloadFile.java");
			}
			//log that the download failed
			logger.error("Error trying download "+remoteFiles+" to ["+localFolder+"].");
		}
		//notify the mainframe that it is done	
		frame.changeTransitFile(remoteFile, false);
		//notify the log that we are done downloading
		logger.info("End download");
	}
}
