package edu.harvard.mgh.purcell.gplink.data;

import org.apache.log4j.Logger;
import org.w3c.dom.*;
import org.xml.sax.*;

import com.sshtools.common.authentication.KBIRequestHandlerDialog;
import com.sshtools.common.authentication.PasswordAuthenticationDialog;
import com.sshtools.common.hosts.AbstractHostKeyVerification;
import com.sshtools.j2ssh.SshClient;
import com.sshtools.j2ssh.authentication.AuthenticationProtocolException;
import com.sshtools.j2ssh.authentication.AuthenticationProtocolState;
import com.sshtools.j2ssh.authentication.KBIAuthenticationClient;
import com.sshtools.j2ssh.authentication.PasswordAuthenticationClient;
import com.sshtools.j2ssh.transport.InvalidHostFileException;
import com.sshtools.j2ssh.transport.TransportProtocolException;

import javax.swing.JOptionPane;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

import edu.harvard.mgh.purcell.gplink.mainGUI.*;
import edu.harvard.mgh.purcell.gplink.base_form.GeneralUtil;
/**
 * This contains the data for using our operation traking.
 * @author Kathe Todd-Brown
 *
 */
public class Project extends KeyWords{
	
	private static Logger logger = Logger.getLogger(Project.class);
	//private Logger superlogger = Logger.getLogger("edu.harvard.mgh.purcell.gplink");
	
	
	///////////////////////////////////////////////////////
	///////////////Non-Static Variables ///////////////////
	///////////////////////////////////////////////////////

	/**
	 * contains all the file and operations information
	 */
	private Document projectInfo;
	
	/**
	 * contains the root folder this is all in
	 */
	private File rootfolder;
	
	/**
	 * path to the project folder in the remote machine
	 */
	private String remotefolder;
	
	/**
	 * remote host
	 */
	private String host;
	
	/**
	 * remote user
	 */
	private String user;
	
	/**
	 * port for ssh connection
	 */
	private int port;

	/**
	 * is the session flagged as read only
	 */
	private boolean readonly;
		
	/**
	 * ssh connection
	 */
	private SshClient conn;
	
	/**
	 * frame that we are interacting with the project with
	 */
	public MainFrame frame;
	
	/**
	 * automatically checks for new files
	 */
	public AutoUpdater gplinkLog;
	
	
	///////////////////////////////////////////////////////
	///////////////  Getter Functions  ////////////////////
	///////////////////////////////////////////////////////

	private String getUser(){
		return user;
	}
	
	public SshClient getConn(){
		return conn;
	}
	
	public String getHost(){
		return host;
	}
	
	public int getPort(){
		return port;
	}
	
	public boolean getBrowse(){
		return readonly;
	}
	
	public String getUpdaterIncrament(){
		return projectInfo.getDocumentElement().	
			getAttribute(KeyWords.UPDATE_KEY);
	}
	
	/**
	 * get the metafile version
	 * @return the version of the current project
	 */
	public String getVersion(){
		return projectInfo.getDocumentElement().getAttribute(KeyWords.VS_KEY);
	}
	
	/**
	 * get the user defined editor command
	 * @return a string identifying the current user
	 * defined editor
	 */
	public String getEditor(){
		return projectInfo.getDocumentElement().getAttribute(USER_EDITOR_KEY);
	}
	
	/**
	 * get the default editor associated with the os type
	 * @return a string contianing the default editor command
	 */
	public String getDefaultEditor(){
		if(System.getProperty("os.name").contains("window")
				||  System.getProperty("os.name").contains("Window")){
			//you use window's you poor poor baby
			return WINDOW_DEFAULT_EDITOR;
		} else if(System.getProperty("os.name").contains("mac")
				|| System.getProperty("os.name").contains("Mac")){
			return MAC_DEFAULT_EDITOR;
		} 	
		else
			return LINUX_DEFAULT_EDITOR;
	}
	
	/**
	 * return the current path of plink
	 * @return returns a string that contains the location of plink
	 */
	public String getPlinkPath(){
		return projectInfo.getDocumentElement().getAttribute(PLINK_PATH_KEY);
	}
	
	/**
	 * Get the full haploview path name
	 * @return a string contianing the location of haploview
	 */
	public String getHaploviewPath(){
		return projectInfo.getDocumentElement().getAttribute(HAPLOVIEW_PATH_KEY);
	}
	
	/**
	 * Get the string that is appended to haploview java launch command
	 * in this project
	 * @return return a string that should be appended to the haploview
	 * launching command
	 */
	public String getHaploviewAppend(){
		return projectInfo.getDocumentElement().getAttribute(HAPLOVIEW_APPEND);
	}
	
	/**
	 * look at the root
	 * @return the root directory folder
	 */
	public File getRoot(){
		return rootfolder;
	}
	
	/**
	 * get the string that defines the remote project file
	 * @return the path of the remote project directory
	 */
	public String getRemoteFile(){
		return remotefolder;
	}
	
	
	
	/**
	 * Get any prefix to that belongs before the plink command is passed
	 * @return a string that is the plink prefix
	 */
	public String getPlinkPrefix(){
		return projectInfo.getDocumentElement().getAttribute(PLINK_PREFIX);
	}
		
	/**
	 * get the name of a project
	 * @return a String which is the name of the project
	 */
	public String getProjectName(){
		if (projectInfo == null)
			return "";
		return projectInfo.getDocumentElement().getAttribute(NAME_KEY);
	}
	
	///////////////////////////////////////////////////////
	///////////////  Setter Functions  ////////////////////
	///////////////////////////////////////////////////////

	/**
	 * Set the timing incraments
	 * @param time_sec is a string that holds the number of seconds
	 * between each auto-updater or NULL if there is no updater set
	 */
	public void setUpdaterIncrament(String time_sec){
		projectInfo.getDocumentElement().setAttribute(
				KeyWords.UPDATE_KEY, time_sec);
	}
	
	/**
	 * Set the browse flag
	 * @param browse a boolean that states the if you want the session
	 * to be browse only (true) or fully accessed (false)
	 */
	public void setBrowse(boolean browse){
		if(frame != null
				&& frame.menuStuff != null){
			frame.menuStuff.setMenuLock(!browse);
		}
		readonly = browse;
		
		//don't close the autoupdater because then we 
		//...loose the icon's for the optree
	}
	
	/**
	 * Set the current host.
	 * @param h a string discribing the host name on a remote
	 * session
	 */
	public void setHost(String h){
		host = h;
	}
	
	
	/**
	 * Set the current user.
	 * @param u a string discribing the user name on the remote
	 * session
	 */
	public void setUser(String u){
		user = u;
	}
	
	/**
	 * Set the port number.
	 * @param p an int specifying the port number for a remote ssh
	 * connection
	 */
	public void setPort(int p){
		port = p;
	}
	
	/**
	 * Set the Connection for the project.
	 * @param c the connection that will be used in the 
	 * remote connection
	 */
	public void setConnection(SshClient c){
		conn = c;
		//hasBroken = false;
	}

	/**
	 * Set the local project file without initalizing the
	 * entire project.
	 * @param given file that makes the local project file
	 */
	public void setLocalRoot(File given){
		rootfolder = given;
		/*if(given != null){
			try {
				superlogger.addAppender(new FileAppender(
						new PatternLayout("%r [%t] %-5p %c{3} - %m%n"), 
						new File(rootfolder, "gPLINK_debugger.txt").getAbsolutePath(),
						false));
			} catch (IOException e) {
				e.printStackTrace();
			}
		}*/
	}
	
	/**
	 * set the path of plink
	 * @param p absolute path name of the plink exicutable
	 */
	public void setPlinkPath(String p){
		projectInfo.getDocumentElement().setAttribute(PLINK_PATH_KEY, p);
	}
	
	/**
	 * Set the haploview path
	 * @param p absolute path name of the haploview .jar file
	 */
	public void setHaploviewPath(String p){
		projectInfo.getDocumentElement().setAttribute(HAPLOVIEW_PATH_KEY, p);
	}
	
	/**
	 * Set an append to the command call to haploview
	 * @param s a string to append to haploview call
	 */
	public void setHaploviewAppend(String s){
		projectInfo.getDocumentElement().setAttribute(HAPLOVIEW_APPEND, s);
	}
	
	/**
	 * Set a preference to the plink command call
	 * @param p a string to prefix to the plink call
	 */
	public void setPlinkPrefix(String p){
		projectInfo.getDocumentElement().setAttribute(PLINK_PREFIX, p);
	}
	
	/**
	 * Set the user defined editor. This is a command to open an editor
	 * @param s an editor command
	 */
	public void setEditor(String s){
		projectInfo.getDocumentElement().setAttribute(USER_EDITOR_KEY, s);
	}
	
	/**
	 * If a connection is open then reset the remote project file.
	 * @param localFile a local file that reflects the remote file
	 * @param remoteFile an absolute path to the remote project file
	 */
	public void setRemoteFile(File localFile, String remoteFile){
		logger.info("(setRemoteFile) Start setting the remote project.");		
		
		if(localFile == null
				&& remoteFile.equals("")){
			rootfolder = null;
			remotefolder = "";
			logger.info("(setRemoteFile) Setting null and empty values for remote project");
			return;
		}
		
		//check that the connection is there
		if(conn == null){
			new ErrorDialog(frame, "No connection is initalized " +
					"therefor I can not set the remote project file. \n" +
			"setRemoteFile@Project.java");
			logger.error("(setRemoteFile) Can not set the remote project with a null connection.");
			return; 
		}
		
		//just set rootfile don't do anything else
		setLocalRoot(localFile);
		
		//reset the remote project
		remotefolder = remoteFile;
		
		////////////////////////////////////
		////// Save the remote project configuration to local project
		////////////////////////////////////
		logger.info("(setRemoteFile) Save the remote configurations.");
		Properties config = new Properties();
		
		try{
			config.setProperty(REMOTE_PROJECT_KEY, remotefolder);
			config.setProperty(USER_KEY, getUser());
			config.setProperty(HOST_KEY, getHost());
			config.setProperty(PORT_KEY, new Integer(getPort()).toString());
			config.store(new FileOutputStream(new File(getRoot(), REMOTE_FILE)), 
			"gPLINK remote configuration");
		} catch (IOException e) {
			new ErrorDialog(frame, "Error saving the remote configuration file. \n" +
			"setRemoteFile@Project.java");
		}

		/////////////////////////////////
		////// load the data from the metafile
		/////////////////////////////////
		logger.info("(setRemoteFile) Load the metafile");
		loadData();
		
		/////////////////////////////////
		////// Start the auto updater
		/////////////////////////////////
		logger.info("(setRemoteFile) Start the auto updater.");
		//if it's attached to a MainFrame
		if(frame != null){
			//and there's an active gplinkLog
			if(gplinkLog != null)
				//cancel the gplinkLog
				gplinkLog.cancel();
			//reset the gplinkLog
			gplinkLog = new AutoUpdater(this);
		} else {
			//otherwise set the log to null since there's
			//no need to update a null MainFrame
			gplinkLog = null;
		}
		
		//////////////////////////////////
		////// Deal with locking
		/////////////////////////////////
		logger.info("(setRemoteFile) Dealing with locking");
		
		//if the lock is avalible and we aren't on
		//...browse then take the lock
		if(avalibleLock() && getBrowse() == false){
			lockMeta();
		}
		
		if(avalibleLock() == false
				&& hasMetaLock() == false 
				&& getBrowse() == false){
			int choice = JOptionPane.showConfirmDialog(frame, 
					"Some other instance of gPLINK has the lock. " +
					"Would you like to steal the " +
					"lock to save the project information?\n " +
					"If you don't then you will be unable to add any new operations. ", 
					"Steal?", JOptionPane.YES_NO_OPTION);
			if(choice == JOptionPane.CLOSED_OPTION
					|| choice == JOptionPane.NO_OPTION){
				setBrowse(true);
			}
			else
				lockMeta();
		}
		

	}
	
	/**
	 * Check to see if the lock is avalable for this project.
	 * A lock is unavaliable if the 1) the project is in browse
	 * mode 2) a lock code that does not belong to this project
	 * is writen to the lock file.
	 * <p>This function uses "less" to look at remote files.
	 * @return true if the lock is avaliable
	 */
	public boolean avalibleLock(){
	
		//if the browse is set then we don't have
		//...permision to even consider the lock
		if(getBrowse() == true){
			logger.info("(avalibleLock) Browse mode active. No lock avaliable.");
			return false;
		}
		
		/////////////////////
		//////lock for remote projects
		/////////////////////
		if(isRemote()){
			//use less to peek at the lock file
			String cmd = "less " + GeneralUtil.makeSpacesHard(getRemoteFile() 
					+ "/" + LOCK_FILE);
			SendCommand getLock = new SendCommand(cmd, this);
			getLock.execute();
			Vector <String> output = getLock.outputLines;
			String remoteLock = null;
			if(output != null && output.size() >= 1)
					remoteLock = output.get(0);	
			logger.info("(avalibleLock) Current key is: "+LOCKED_CODE
					+"Remote lock reads: ["+remoteLock+"]");
			//if the file had a lock code in it
			//...that does not belong to the project
			//...then return false
			if(remoteLock != null
					&& !remoteLock.equals(LOCKED_CODE) 
					&& remoteLock.contains(LOCKED_BASE))
				return false;
			else
				return true;
		}
		
		////////////////////////////
		///////lock for non-remote projects
		////////////////////////////
		try {
			//if the file doesn't exist create a new lock file
			new File(getRoot(), LOCK_FILE).createNewFile();
		} catch (IOException e) {
			logger.error("(avalibleLock) Error trying to create a local lock file!");
			new ErrorDialog(frame, "Error ensuring local lock file exists.\n " +
					"avaliableLock@Project.java");
			//we can't lock a folder we can't create a lock file in
			return false;
		}
		
		///////
		//read the local lock file
		//////
		String localLock = "";
		try{
			BufferedReader in = new BufferedReader(
					new FileReader(new File(getRoot(), LOCK_FILE)));
			//the lock code is on the first line and this is all
			//...that's in the file
			localLock = in.readLine();
			in.close();
		}catch(IOException e){
			logger.error("(avalibleLock) Can not read the local lock file!");
			new ErrorDialog(frame, "Error trying to read in local lock.\n" +
					"avalibleLock@Project.java");
			return false;
		}
		
		logger.info("(avalibleLock) Current key is: "+LOCKED_CODE
				+"Remote lock reads: ["+localLock+"]");
		
		//make sure the lock isn't null (should never happen)
		//...if the lock isn't equal to the current locked_code
		//...but contains the lock base then the lock isn't avaliable
		if (localLock != null
				&& !localLock.equals(LOCKED_CODE) 
				&& localLock.contains(LOCKED_BASE))
			return false;
		else
			return true;
		
		
	}
	
	/**
	 * Check to see if the project currently has the lock
	 * <p>"less" is used to look at the remote project lock
	 * @return true if the lock key equals that of the current project
	 */
	public boolean hasMetaLock(){
		
		//if the browse mode is on then we don't even
		//...consider the metafile
		if(getBrowse() == true)
			return false;
		
		///////////////////////
		//////Check the remote lock
		///////////////////////
		if(isRemote()){
			String cmd = "less " + GeneralUtil.makeSpacesHard(getRemoteFile() 
					+ "/" + LOCK_FILE);
			SendCommand getFiles = new SendCommand(cmd, this);
			getFiles.execute();
			String remoteLock = null;
			Vector <String> output = getFiles.outputLines;
			if(output != null
					&& output.size() >= 1)
				remoteLock = output.get(0);
			logger.info("(hasMetaLock) Current key is: "+LOCKED_CODE
					+"Remote lock reads: ["+remoteLock+"]");
			if(remoteLock == null
					|| !remoteLock.equals(LOCKED_CODE))
				return false;
			else
				return true;
		}
		
		//////////////////////
		//////Check the local lock
		/////////////////////
		
		//Make sure that there is a file to look at
		try { 
			//if we sucessfuly create a lock file then obviously
			//...this project can not have the lock
			if(new File(getRoot(), LOCK_FILE).createNewFile()){
				return false;
			}
		} catch (IOException e) { 
			logger.error("(hasMetaLock) Error trying to create a local lock file!");
			new ErrorDialog(this, "Error trying to create the local lock.\n" +
				"hasMetaLock@Project.java");
			return false;
		}
		
		//read in the lock
		String localLock ="";
		try{
			BufferedReader in = new BufferedReader(
					new FileReader(new File(getRoot(), LOCK_FILE)));
			localLock = in.readLine();
			in.close();
		}catch(IOException e){
			logger.error("(hasMetaLock) Error trying to read a local lock file!");
			new ErrorDialog(this, "Error trying to read local project lock.\n" +
					"hasMetaLock@Project.java");
			return false;
		}
		
		logger.info("(hasMetaLock) Current key is: "+LOCKED_CODE
				+"Remote lock reads: ["+localLock+"]");
		
		//return the match to the local lock
		return localLock != null && LOCKED_CODE.equals(localLock);
	}

	
	/**
	 * Lock the project home directory
	 *  so that no other instance of gPLINK will conflict.
	 * If the project is remote then "echo" is used to 
	 * write to the lock file.
	 */
	public void lockMeta(){
		logger.info("(lockMeta) Write to lock file: "+LOCKED_CODE);
		
		////////////////////
		/////// lock the remote directory
		////////////////////
		if( isRemote()){
			//use echo to write the lock code to the file
			String cmd = "echo \"" + LOCKED_CODE + "\" > " 
			+ GeneralUtil.makeSpacesHard(getRemoteFile()+"/" +LOCK_FILE);
			new SendCommand(cmd, this).execute();
			return;
		}
		
		///////////////////
		////// try to lock the local project file
		///////////////////
		try {
			//write the lock code to the local file
			BufferedWriter out = new BufferedWriter(
					new FileWriter(new File(getRoot(), LOCK_FILE)));
			out.write(LOCKED_CODE);
			out.newLine();
			out.close();
		} catch (IOException e) {
			logger.error("(lockMeta) Error writing to the local lock file.");
			new ErrorDialog(frame, "Error writing the lock file locally. \n" +
			"lockMeta@Project.java");
			return;
		}
	}
	
	/**
	 * Unlock the project file so that no other instance of gPLINK will confict.
	 * Remote projects use "echo" to write the lock file.
	 *
	 */
	public void unlockMeta(){
		
		logger.info("(lockMeta) Write to lock file: " + UNLOCKED_CODE);
		
		///////////////////
		/////// get the lock from the remote
		///////////////////
		if( isRemote()){
			String cmd = "echo \"" + UNLOCKED_CODE + "\" > " 
			+ GeneralUtil.makeSpacesHard(getRemoteFile()+"/" + LOCK_FILE);
			new SendCommand(cmd, this).execute();
			return;
		}
		
		////////////////////
		/////// try to lock the local project file
		////////////////////
		try {
			BufferedWriter out = new BufferedWriter(
					new FileWriter(new File(getRoot(), LOCK_FILE)));
			out.write(UNLOCKED_CODE);
			out.newLine();
			out.close();
		} catch (IOException e) {
			logger.error("(lockMeta) Error writing to the local lock file.");
			new ErrorDialog(frame, "Error writing the lock file locally. \n" +
			"unlockMeta@Project.java");
		}
	}
	
	
	
	///////////////////////////////////////////////////////
	/////////////// Remote Functions   ////////////////////
	///////////////////////////////////////////////////////
	
	
	/**
	 * Close the remote connection and clear the remote project file parameter.
	 */
	public void closeRemote(){
		//if the connection is not null then close it
		if(conn != null)
			conn.disconnect();
		//make sure the connection is null
		conn = null;
		//clear the remote project file parameter
		setRemoteFile(null, "");
		
		//TODO why is this here??
		//set the browse
		//setBrowse(frame.menuStuff.browseOnly.isSelected());
	}
	
	/**
	 * This is a host key validator that automatically accepts any
	 * host key sent but doesn't actually save any of them.
	 * @author Kathe Todd-Brown
	 *
	 */
	private class myHostKeyValidator extends AbstractHostKeyVerification{
		public myHostKeyValidator() throws InvalidHostFileException {super();}
		@Override
		public void onDeniedHost(String host) throws TransportProtocolException {
			new ErrorDialog(frame, "Access to '" + host + "' is denied.\n" +
                    "Verify the access granted/denied in the allowed hosts file.");
		}
		@Override
		public void onHostKeyMismatch(String host, String allowedHostKey, 
				String actualHostKey) throws TransportProtocolException {
			allowHost(host, actualHostKey, false);}
		@Override
		public void onUnknownHost(String host, String hostKeyFingerprint) 
		throws TransportProtocolException {
			allowHost(host, hostKeyFingerprint, false);}
	}
	
	/**
	 * Connect the ssh client.
	 * @return false if an error that we can cetch happened (not always the case!)
	 * true otherwise.
	 */
	@SuppressWarnings("unchecked")
	public boolean connect(){
		//if the connection isn't null
		if(conn != null)
			//be sure that we are disconnected
			conn.disconnect();
		//set the connection to something new
		conn = new SshClient();
			
		////////////
		///Establish connection
		////////////
		try {
			conn.connect(host,port, new myHostKeyValidator());
		} catch (NumberFormatException e) {
			logger.error("(connect) Can not establish a connection: NumberFormatException");
			new ErrorDialog(frame,". Error trying to establish connection.\n" +
					"connect:Project.java");
			return false;
		} catch (IOException e) {
			logger.error("(connect) Can not establish a connection: IOException");
			new ErrorDialog(frame, "Error trying to establish " +
					"ssh connection " +
					"["+ user+"@"+host +":"+ port +"]. \n" +
					"connect:Project.java");
			return false;
		}
		
		///////////////
		/////authenticate the connection
		///////////////
		
		//flags if we have a key-board interactive
		boolean kyboardI = false;
		try {
			List <String> options = conn.getAvailableAuthMethods(user);
		
			String allOptions = "";
			for(String name: options){allOptions += " ["+ name + "]";}
			logger.info("(connect) We have the following auterization options:"+allOptions);
			//if there is a keyboard interactive option
			kyboardI = options.contains("keyboard-interactive");
		} catch (IOException e2) {
			logger.error("(connect) Can not establish ssh auth methods");
			new ErrorDialog(frame, "Error trying to establish " +
					"ssh auth methods" +
					"["+ user+"@"+host +":"+ port +"]. \n" +
					"connect():Project.java");
			return false;
		}
		
		//by detault we fail the autherization
		int result = AuthenticationProtocolState.FAILED;
		
		//create authenticators and authenticate the connection
		if(kyboardI){
			KBIAuthenticationClient kbi =
				new KBIAuthenticationClient();
			KBIRequestHandlerDialog dialog = 
	              new KBIRequestHandlerDialog(frame);
			kbi.setKBIRequestHandler(dialog);
			kbi.setUsername(user);
			
			//get the result
			try {
				result = conn.authenticate(kbi);
			} catch (IOException e) {
				logger.error("(connect) Can not authenticate " +
						"(keyboard-interative) ssh connection " +
						"["+ user + "@" +host+":"+ port+"]. \n");
				new ErrorDialog(frame, "Error trying to authenticate " +
						"(keyboard-interative) ssh connection " +
						"["+ user + "@" +host+":"+ port+"]. \n" +
						"launchLogin:ProjectDialog.java");
				return false;
			}
		} else {
			PasswordAuthenticationClient pwd = 
				new PasswordAuthenticationClient();
			PasswordAuthenticationDialog dialog =
				new PasswordAuthenticationDialog(frame);
			pwd.setUsername(user);
			
			try {
				pwd.setAuthenticationPrompt(dialog);
			} catch (AuthenticationProtocolException e) {
				logger.error("(connect) Can not set authenticate prompt" +
						"(password)" +
						"["+ user + "@" +host+":"+ port+"]. \n");
				new ErrorDialog(frame, "Can not set authenticate prompt" +
						"(password)" +
						"["+ user + "@" +host+":"+ port+"]. \n" +
						"launchLogin:ProjectDialog.java");
				return false;
			}
			try {
				result = conn.authenticate(pwd);
			} catch (IOException e) {
				logger.error("(connect) Can not authenticate " +
						"(password) ssh connection " +
						"["+ user + "@" +host+":"+ port+"]. \n");
				new ErrorDialog(frame, "Error trying to authenticate " +
						"(password) ssh connection " +
						"["+ user + "@" +host+":"+ port+"]. \n" +
						"launchLogin:ProjectDialog.java");
				return false;
			}
		}
		
		/////////////////
		////// Check the results.
		/////////////////
	  
		if(result==AuthenticationProtocolState.FAILED){
			logger.error("(connect) Error failure to " +
					"establish ssh connection"+
					"["+ user + "@" +host +":"+ port+"].");
			new ErrorDialog(frame, "Error failure to " +
					"establish ssh connection"+
					"["+ user + "@" +host +":"+ port+"]. \n" +
					"connect:Project.java");
			return false;
		}
		if(result==AuthenticationProtocolState.PARTIAL){
			logger.error("(connect) Error failure to " +
					"establish full ssh connection"+
					"["+ user + "@" +host +":"+ port+"].");
			new ErrorDialog(frame, "Error failure to " +
					"establish full ssh connection"+
					"["+ user + "@" +host +":"+ port+"]. \n" +
					"connect:Project.java");
		    return false;
		}
		
		if(result==AuthenticationProtocolState.COMPLETE){
			logger.info("(connect) Auth complete!");
			return true;
		}
		
		logger.error("(connect) We shouldn't be here!");
		//we don't know what happened if we get here but we don't like it!
		return false;
	}
	
	/**
	 * Re-establish a lost connection.
	 * @return false if the user choses not or there was a problem
	 * we caught establishing the connection (note that we don't
	 * cetch all the problems!), true otherwise
	 */
	public boolean connectionLost(){

		int choice = JOptionPane.showConfirmDialog(frame, 
			"There is a problem with the connection.\n" +
			"Would you like try to reconnect?",
			"Connection Error",
		    JOptionPane.YES_NO_OPTION);
		
		if(choice == JOptionPane.YES_OPTION){
			return connect();
		}else{
			//if we can't establish a connection then we
			//can only browse what's in the local file
			setBrowse(true);
			//set the connection to null
			conn = null;
			return false;
		}
	}
	
	//TODO this method is decrepid and should be deleted
	/**
	 * A project is flagged as remote if there is a 
	 * remote config file in the local directory.
	 * @return true if it is flagged and false otherwise
	 */
	public boolean isflaggedRemote(){
		if(getRoot() == null 
				|| new File(getRoot(), REMOTE_FILE).exists() == false)
			return false;
		else
			return true;
	}
	
	/**
	 * A project is remote if it has a non-empty local project folder
	 * a non-null connection and has a config file for a remote project
	 * @return true if the project is remote, false otherwise
	 */
	public boolean isRemote(){
		if(getRoot() == null 
				|| conn == null
				|| new File(getRoot(), REMOTE_FILE).exists() == false)
			return false;
		else
			return true;		
	}
	
	/**
	 * Set the root file we are working in and load the new data
	 * file that we find there. Checks also to see that the given
	 * file is a directory. Note that this local root must be set
	 * before the remote project can be set.
	 * @param f this is the folder that we are working from
	 */
	public void setLocalProject(File f){
		logger.info("(setLocalProject) Entering set Local project");
		/////////////////
		/////make sure everything is cleaned up
		/////////////////
		
		//if the connection is open we want close it
		if(conn != null)
			closeRemote();
		
		//take out the file that flags this as remote
		new File(f, REMOTE_FILE).delete();
		
		//////////////
		///// cetch the null case
		//////////////
		
		//if we want to set the 
		if(f == null){
			//set the project info to null
			projectInfo = null;
			//set the root file to null
			rootfolder = null;
			//and there's an active gplinkLog
			if(gplinkLog != null)
				//cancel the gplinkLog
				gplinkLog.cancel();
			//we are done
			return;
		}
		
		//////////////
		///// initalize stuff
		//////////////
		
		//set the folder
		setLocalRoot(f);
		//rootfolder = f;	
		
		//load the project information
		loadData();
		
		//check that it is a valid folder and not write protected
		if(f.isDirectory() == true && f.canWrite() == true){
			/////////
			/////start a new auto updater
			/////////
			
			//if it's attached to a MainFrame
			if(frame != null && projectInfo != null){
				//and there's an active gplinkLog
				if(gplinkLog != null)
					//cancel the gplinkLog
					gplinkLog.cancel();
				//reset the gplinkLog
				gplinkLog = new AutoUpdater(this);
			} else {
				//otherwise set the log to null since there's
				//no need to update a null MainFrame
				gplinkLog = null;
			}
			
			///////////
			////take care of the locking
			///////////
			
			if(avalibleLock() && getBrowse() == false){
				lockMeta();
			}
			
			if(avalibleLock() == false
					&& hasMetaLock() == false 
					&& getBrowse() == false){
				int choice = JOptionPane.showConfirmDialog(frame, 
						"Some other instance of gPLINK has the lock. " +
						"Would you like to steal the " +
						"lock to save the project information?\n " +
						"If you don't then you will be unable to add any new operations. ", 
						"Steal?", JOptionPane.YES_NO_OPTION);
				if(choice == JOptionPane.CLOSED_OPTION
						|| choice == JOptionPane.NO_OPTION){
					setBrowse(true);
				}
				else
					lockMeta();
			
			}
		} else{
			logger.error("(setLocalProject) Cannot set the local project to a non-directory or" +
					" non-writable folder.");
			new ErrorDialog(frame,"Cannot set the local project to a non-directory or" +
					" non-writable folder.\n setLocalProject()@Project.java");
		}
		
		logger.info("(setLocalProject) Exiting");
		
	}
	
	
	/**
	 * Get a list of files in the current project directory, local or remote.
	 * <p> The remote option uses "ls".
	 * @return a vector of strings that hold the full path names of all the files
	 * in the project folder.
	 */
	public Vector <String> getFileList(){
		Vector <String> ans = new Vector<String>();
		
		//return an empty vector if the local root is null
		if(getRoot() == null){
			return ans;
		}
		
		//
		if(isRemote() == true){
			String cmd = "ls " + GeneralUtil.makeSpacesHard(remotefolder);
			SendCommand getFiles = new SendCommand(cmd, this);
			getFiles.execute();
			ans = getFiles.outputLines;
		}else{
			File [] temp = getRoot().listFiles();
			for(int i = 0; i < temp.length; i ++){
				ans.add(temp[i].getAbsolutePath());
			}
		}
		return ans;
	
	}
	
	/**
	 * get the command line for a given operation
	 * @param name the name of the operation you are looking for
	 * @return the command line attached to that operation
	 */
	public String getOpCommand(String name){
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right name
				if(((Element)index).getAttribute(NAME_KEY).equals(name))
					return ((Element)index).getAttribute(CLINE_KEY);
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		
		return null;
	}
	
	/**
	 * get the discription of an operation based on a given name
	 * @param name a string containing the name of the operation
	 * @return a string contianing the discription associated with
	 * the operation
	 */
	public String getOpDiscription(String name){
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type (text, entity or operation)
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right file/operation name
				if(((Element)index).getAttribute(NAME_KEY).equals(name)){
					return index.getFirstChild().getNodeValue();
				}
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		
		return null;
	}
	
	/**
	 * get a list of all the operation names in the project
	 * @return a vector of all operations names this will be
	 * empty if there are no operations
	 */
	public Vector<String> getAllOpName(){
		//if we don't have any document initalized then logically we shouldn't
		//...return any operations
		if (projectInfo == null)
			return null;
		//create an answer for us to return
		Vector<String> ans = new Vector<String>();
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you fall off
		while(index != null){
			//checks to see if we are at the right type (text, entity or operation)
			if(index.getNodeName().equals(OP_KEY)){
				//grab the name, add it to your answer
				ans.add(((Element) index).getAttribute(NAME_KEY));
			}//end if
			//go to the next sibling
			index = index.getNextSibling();
		}//end while
		//return the ans
		return ans;
	}
	
	/**
	 * returns a list of the files used to construct a given operation
	 * and their corrisponding discriptions.
	 * @param opName the operation name that you are looking for
	 * @param inOut flag that looks at either the INFILE or OUTFILE of
	 * the operation
	 * @return a list of file nodes and text (used form XML formating) nodes
	 */
	public String [] [] getOpFiles(String opName, String inOut){
		String [][] ans = null;
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you fall off
		while(index != null){
			//checks to see if we are at the right type 
			if(index.getNodeName().equals(OP_KEY)){
				//we know that all the operations are elements
				Element temp = (Element) index;
				//check the name
				if(temp.getAttribute(NAME_KEY).equals(opName)){
					//go though the children nodes
					Node index2 = index.getFirstChild();
					while(index2 != null){
						//find the infile/outfile node
						if(index2.getNodeName().equals(inOut)){
							//get the children some of which will be text nodes
							//and some of which will be file nodes
							//int numFiles = Integer.parseInt(
							//		((Element)index2).getAttribute(NUM_FILE_KEY));
							//keep track of the index we are on so that we
							//...can populate the array
							int curIndex = 0;
							//set the answer finally
							Vector <String> fileNames = new Vector<String>();
							Vector <String> fileDisc = new Vector<String>();
							//ans = new String [2] [numFiles];
							//go through the childre nodes of the file node
							Node index3 = index2.getFirstChild();
							while (index3 != null){
								//make sure that it is a file and 
								//...not a text node
								if(index3.getNodeName().equals(FILE_KEY)){
									//ans[0][curIndex] = 
									fileNames.add(
										((Element)index3)
										.getAttribute(NAME_KEY));
									//ans[1][curIndex] = 
									fileDisc.add(
										index3.getTextContent());
									curIndex ++;
								}
								//get the next possible file node
								index3 = index3.getNextSibling();
							}
							ans = new String[2][fileNames.size()];
							for(int i = 0; i < ans[0].length; i ++){
								ans[0][i] = fileNames.get(i);
								ans[1][i] = fileDisc.get(i);
							}
						
							
							//we went though all the files and can now quit
							return ans;
						}//end if: outfile/infile name check
						
						//go to the next child node of our operation
						index2 = index2.getNextSibling();
					}//end while: going though all second tier children
					
					//we checked all decendents of our matching operation
					return ans;
				}//end if: operation name check
			}//end if: operation type check
			
			//go to the next sibling
			index = index.getNextSibling();
		}//go though all first tier children
		
		return ans;
	}

	
	/**
	 * construct the project by either settting&reading the root directory or
	 * setting all the internal info to null.
	 * @param f diretory that we wish to point the root to
	 */
	public Project(File f, MainFrame mf){
		setBrowse(false);
		frame = mf;
		
		//if the project is initalized with info
		if (f != null)
			//set the root folder we are working in
			setLocalProject(f);
		else{
			//set everything to null
			rootfolder = null;
			projectInfo = null;
		}
	}

	/**
	 * Save, unlock and set the root file to null.
	 *
	 */
	public void closeOldProject(){
		if(gplinkLog != null){
			gplinkLog.cancel();
			gplinkLog = null;
		}
		
		if(projectInfo != null
				&& getRoot() != null){
			writeProjectInfo(true, false, true);
			
			if(hasMetaLock())
				unlockMeta();
			
			projectInfo = null;
		}	
		
		frame.folderViewer.clearViewer();
		
		setBrowse(frame.menuStuff.browseOnly.isSelected());
			
		closeRemote();
		
		setLocalProject(null);
		
	}
	
	/**
	 * Update all the output files in the project
	 *
	 */
	public void updateFiles(){
		updateFiles(false);
	}
	/**
	 * Update all the output file attached to the operations
	 * check to see if there are any new files.
	 * @param forceUpdate is a boolean flag that forces another
	 * download of the log files
	 */
	public void updateFiles(boolean forceUpdate){
		//if we don't actually have a project then we don't have anything to do
		if(projectInfo == null)
			return;
		
		//a vector that holds all the operation names
		Vector <String> allLogRoots = new Vector <String>();
		//a vector that holds all the files in the home folder
		Vector <String> allFiles = getFileList();
		//a vector that holds all the files ot be downloaded
		Vector <String> filesToDownload = new Vector<String>();
		
		////////////////////////
		//////go through and pull the operation names
		//////from the log files in the folder and
		//////the files that need to be downloaded
		////////////////////////
		for(String files: allFiles){
			//if the file is a log or gplink file
			if(files.endsWith(".log") 
					|| files.endsWith(OP_STATUS_EXT)){
				//pull out the short file name
				String fileName = GeneralUtil.fileName(files);
				if(files.endsWith(".log")){
					//chop off the log bit and use it as an
					//...operation name
					allLogRoots.add(fileName.substring(0, fileName.length()-4));
				}
				//if it's a remote project
				if( this.isRemote() 
						//and the file doesn't exist or updates forced
						&& ( ! new File(getRoot(), fileName).exists()
						|| forceUpdate == true)){
					//put it in the queue to download
					filesToDownload.add(remotefolder + "/" + files);
				}
			}
		}
		
		//if it's remote and they have files queued to 
		//...download then download the files
		if(isRemote() && !filesToDownload.isEmpty()){
			DownloadFile getFiles = new DownloadFile(
					 filesToDownload,
					getRoot().getAbsolutePath(), getConn(), 
					frame, false);
			getFiles.start();
		}
		
		///////////////////////////
		//////Go through and d
		
		//start at the first child of the root
		Node opindex = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child (operation)
		while(opindex != null){
			//checks to see if we are at the right type
			if(opindex.getNodeName().equals(OP_KEY)){
				//create a node to try to find the output folder
				Node folderIndex = opindex.getFirstChild();
				//go to the Outfiles node
				while(folderIndex != null
						&& (folderIndex.getNodeName().equals(OUTFILE_KEY) == false )){
					folderIndex = folderIndex.getNextSibling();
				}
				//should'nt get here unless it's a poorly formated operation
				if(folderIndex == null){
					new ErrorDialog(frame, "Error! Can't find outfile node " +
							"while updating files.\n updateFiles@Project.java");
					return;
				}
				//get the operation name of the node you are at
				String opName = ((Element) opindex).getAttribute(NAME_KEY);
				//get the output files in that operation
				String [][] outFiles = getOpFiles(opName, OUTFILE_KEY);
				//////////////////////////////
				//get a list of all files that have the same root
				////////////////////////////
				HashSet<String> filteredFiles = new HashSet <String>();
				for(String file: allFiles){
					file = GeneralUtil.fileName(file);
					//if the root is between a file sperator and "." then it's
					//...attached to the operation
					if(file.startsWith(opName + "."))
					{
						filteredFiles.add(file);
					}
				}
				//deal with the fact that "." are valid in a file name
				Vector<String> temp = new Vector<String>();
				for(String file:filteredFiles){
					//double check that it doesn't match another
					for(String root: allLogRoots){
						if(!opName.equals(root)
								&& root.contains(opName)
								&& new File(file).getName().startsWith(root+"."))
							temp.add(file);
					}
				}
				for(String file:temp){
					filteredFiles.remove(file);
				}
				
				//we start with one file
				if(outFiles[0].length >= 1){
					//compare the two lists
					for(int i = 0; i < outFiles[0].length; i ++){
						if(filteredFiles.contains(outFiles[0][i]))
							filteredFiles.remove(outFiles[0][i]);
					}
				} else {
					System.err.println("Problem updating files.");
				}
				//add the new files
				for(String s: filteredFiles){
					addFile(folderIndex, s, "");
				}
				
			}
			//go to the next operation
			opindex = opindex.getNextSibling();
		}
		
		return;
	}
	
	/**
	 * load the data from the log file if it exists. If it does not exist
	 * then create a head for the project (note that there is no call to write
	 * to the log file in this operation).
	 *
	 */
	@SuppressWarnings("static-access")
	private void loadData(){
		
		//double check that the rootfile is not null
		if(rootfolder == null){
			return;
		}
		
		
		//use xml document builders to create a new document from the xml
		//...log file
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setIgnoringComments(true);
		DocumentBuilder parser = null;
		
		Properties localConfig = new Properties();
		
		//System.out.println("Flag1 loadData()@Project.java " 
		//		+ (new GregorianCalendar()).getTimeInMillis());
		//only stays this setting if you are in a remote folder
		File configFile = new File(System.getProperty("user.dir")
				+ File.separator + GLOBAL_CONFIG_FILE);
		
		try{
			//System.out.println("loadData@Project.java The config file is: " + configFile.getAbsolutePath());
			localConfig.load(new FileInputStream(configFile));
			
		}catch(Exception e1){
			//System.out.println("loadData@Project.java The config file is: " + configFile.getAbsolutePath());
			configFile = new File(System.getProperty("user.home")
					+ File.separator + GLOBAL_CONFIG_FILE);
			try{
				localConfig.load(new FileInputStream(configFile));
			}catch(Exception e2){
				localConfig.setProperty(PLINK_PATH_KEY, "plink");
				localConfig.setProperty(PLINK_PREFIX, "");
				localConfig.setProperty(HAPLOVIEW_PATH_KEY, "");
				localConfig.setProperty(HAPLOVIEW_APPEND, "");
				localConfig.setProperty(USER_EDITOR_KEY, "");
				localConfig.setProperty(PROJECT1, "");
				localConfig.setProperty(PROJECT2, "");
				localConfig.setProperty(PROJECT3, "");
				localConfig.setProperty(PROJECT4, "");
				localConfig.setProperty(BROWSE_KEY, FALSE);
			}
		}
		
		//System.out.println("Flag2 loadData()@Project.java " 
		//		+ (new GregorianCalendar()).getTimeInMillis());
		Properties remoteConfig = new Properties();
		if(isRemote()){
			DownloadFile getConfig = new DownloadFile( GLOBAL_CONFIG_FILE,
						getRoot().getAbsolutePath(), getConn(), frame, true);
			getConfig.start();
			try {
				getConfig.join();
			} catch (InterruptedException e) {
				new ErrorDialog(frame, "Error waiting for configuration download thread. \n" +
						"loadData@Project.java");
				return;
			}
			try{
			remoteConfig.load(new FileInputStream(
					new File(getRoot(), GLOBAL_CONFIG_FILE)));
			
			} catch (IOException e) {
				//it will all get taken care off later we just don't
				//set the remote properties "right" nothing breaks
				if(remoteConfig.getProperty(PLINK_PATH_KEY) == null)
					remoteConfig.setProperty(PLINK_PATH_KEY, "");
				if(remoteConfig.getProperty(PLINK_PREFIX) == null)
					remoteConfig.setProperty(PLINK_PREFIX, "");
			}
			
		}
			
		if(localConfig.getProperty(PROJECT1) == null)
			localConfig.setProperty(PROJECT1, "");
		if(localConfig.getProperty(PROJECT2) == null)
			localConfig.setProperty(PROJECT2, "");
		if(localConfig.getProperty(PROJECT3) == null)
			localConfig.setProperty(PROJECT3, "");
		if(localConfig.getProperty(PROJECT4) == null)
			localConfig.setProperty(PROJECT4, "");
		if(localConfig.getProperty(HAPLOVIEW_PATH_KEY) == null)
			localConfig.setProperty(HAPLOVIEW_PATH_KEY, "");
		if(localConfig.getProperty(HAPLOVIEW_APPEND) == null)
			localConfig.setProperty(HAPLOVIEW_APPEND, "");
		if(localConfig.getProperty(PLINK_PATH_KEY) == null)
			localConfig.setProperty(PLINK_PATH_KEY, "plink");
		if(localConfig.getProperty(PLINK_PREFIX) == null)
			localConfig.setProperty(PLINK_PREFIX, "");
		if(localConfig.getProperty(USER_EDITOR_KEY) == null)
			localConfig.setProperty(USER_EDITOR_KEY, "");
		if(localConfig.getProperty(BROWSE_KEY) == null)
			localConfig.setProperty(BROWSE_KEY, FALSE);
		
		//System.out.println(config.get(PROJECT1));
		if(!localConfig.get(PROJECT1).equals(rootfolder.getAbsolutePath())){
			if(localConfig.get(PROJECT2).equals(getRoot().getAbsolutePath())){
				localConfig.setProperty(PROJECT2, localConfig.getProperty(PROJECT1));
			}else if(localConfig.get(PROJECT3).equals(getRoot().getAbsolutePath())){
				localConfig.setProperty(PROJECT3, 
						localConfig.getProperty(PROJECT2));
				localConfig.setProperty(PROJECT2, 
						localConfig.getProperty(PROJECT1));
			} else{
				localConfig.setProperty(PROJECT4, 
						localConfig.getProperty(PROJECT3));
				localConfig.setProperty(PROJECT3, 
						localConfig.getProperty(PROJECT2));
				localConfig.setProperty(PROJECT2, 
						localConfig.getProperty(PROJECT1));
			}
				
			localConfig.setProperty(PROJECT1, getRoot().getAbsolutePath());
			
			try{
				localConfig.store(new FileOutputStream(configFile), 
						"gPLINK global configuration");
			}catch(IOException e1){
				new ErrorDialog(frame, "Error storing configuration file.\n" +
						"loadData@Project.java");
				//return;
			}
			
			if(frame != null 
					&& frame.menuStuff != null)
				frame.menuStuff.setPreviousProjects(localConfig);
		}
		
//		create the whole file path of the log file
		File workinglog = new File(rootfolder.getAbsolutePath(), LOG_FILE);
		
		//copy the log file from the remote folder
		if(isRemote()){
			if(workinglog.exists() ){
				//delete any local copy of the metafile we have
				workinglog.delete();
			}
						
			DownloadFile getlog = new DownloadFile(remotefolder 
						+ "/" + LOG_FILE, (getRoot().getAbsolutePath()), getConn(), frame, true);
			getlog.start();
			try {
				getlog.join();
			} catch (InterruptedException e) {
				new ErrorDialog(frame, "Error waiting for metafile download thread. \n" +
				"loadData@Project.java");
				return;
			}
		}
		
		try {
			parser = factory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			new ErrorDialog(frame, "Error trying to create a new document builder.\n" +
					"loadData@Project.java");
			return;
		}
		
		//take the remote parames that are appropreate
		Properties config = localConfig;
		if(isflaggedRemote()){
			config.setProperty(PLINK_PATH_KEY, 
					remoteConfig.getProperty(PLINK_PATH_KEY, "plink"));
			config.setProperty(PLINK_PREFIX, 
					remoteConfig.getProperty(PLINK_PREFIX, ""));
		}
		
		//check that the log file exists
		if(workinglog.exists() ){
			try {
				projectInfo = parser.parse(workinglog);
			} catch (SAXException e) {
				new ErrorDialog(frame, "Error trying to create project, " +
					"can not parse log file.\n Creating new " +
					"project, please import old operations.\n " +
					"loadData@Project.java");
				createNewProject(parser);
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to create project, " +
					"IOException.\n Creating new" +
					"project, please import old operations.\n " +
					"loadData@Project.java");
				createNewProject(parser);
			}
			projectInfo.getDocumentElement().setAttribute(VS_KEY, frame.VS);
		}else{
			createNewProject(parser);
		}
		
		setPlinkPath(config.getProperty(PLINK_PATH_KEY));
		setPlinkPrefix(config.getProperty(PLINK_PREFIX));
		setHaploviewPath(config.getProperty(HAPLOVIEW_PATH_KEY));
		setHaploviewAppend(config.getProperty(HAPLOVIEW_APPEND));
		setEditor(config.getProperty(USER_EDITOR_KEY));
		setBrowse(config.getProperty(BROWSE_KEY).equals(TRUE));
		
		if((!isflaggedRemote() && !projectInfo.getDocumentElement().getAttribute(HOME_KEY).equals(
				rootfolder.getAbsolutePath()))
				|| ( isflaggedRemote() && !projectInfo.getDocumentElement().getAttribute(HOME_KEY).equals(
						getRemoteFile())))
			replaceHome();
		
	}
	
	/**
	 * Create a new project from scratch.
	 * @param parser is a document builder
	 * configuration information for this project
	 */
	@SuppressWarnings("static-access")
	private void createNewProject(DocumentBuilder parser){

		//create a new document
		projectInfo = parser.newDocument();
		
		//put in the header information for the project but no file data
		projectInfo.createTextNode(String.format("%n", (Object []) null));
		Element root = 
			(Element) projectInfo.createElement("project");
		root.setAttribute(NAME_KEY, rootfolder.getName());
		root.setAttribute(VS_KEY, frame.VS);
		//root.setAttribute(EOP_KEY, "0");
		if(!isflaggedRemote()){
			root.setAttribute(HOME_KEY, rootfolder.getAbsolutePath());
		}else{
			root.setAttribute(HOME_KEY, getRemoteFile());			
		}
		
		//apppend the root to the project
		projectInfo.appendChild(root);
		//pad the file with new lines to make it human readable
	    root.appendChild( projectInfo.createTextNode(String.format("%n", (Object []) null)));
	
	    setUpdaterIncrament(UPDATE_DEFAULT);
	}
	
	/**
	 * Check with the user to see if they want the old home file instances
	 * replaces with the new one.
	 *
	 */
	public void replaceHome(){
		String oldHome = projectInfo.getDocumentElement().getAttribute(HOME_KEY);
	
		String newHome;
		String newHomeFolder;
		if(isflaggedRemote()){
			newHome = getRemoteFile();
			newHomeFolder = newHome + "/";
		}else{
			newHome = rootfolder.getAbsolutePath();
			newHomeFolder = newHome + File.separator;
		}
		
		int choice = JOptionPane.showConfirmDialog(null,
				"I detect a different path. " +
				"Do you want to replace all instances of: \n[" +
				oldHome +
				"] with [" + newHome + "]?\n" +
						"(Hint: if you have moved your folder select yes.)", 
				"Replace?", JOptionPane.YES_NO_OPTION);
		if(choice == JOptionPane.NO_OPTION)
			return;
		if(choice == JOptionPane.YES_OPTION){

			Node opIndex = projectInfo.getDocumentElement().getFirstChild();
			//go until you fall off
			while(opIndex != null){
				//checks to see if we are at the right type (text, entity or operation)
				if(opIndex.getNodeName().equals(OP_KEY)){
					Node groupIndex = opIndex.getFirstChild();
					while(groupIndex != null){
						if(groupIndex.getNodeName().equals(INFILE_KEY) 
								||groupIndex.getNodeName().equals(OUTFILE_KEY)){
							Node fileIndex = groupIndex.getFirstChild();
							while(fileIndex != null){
								if(fileIndex.getNodeName().equals(FILE_KEY)){
									String oldName = ((Element) fileIndex).getAttribute(NAME_KEY);
									String tempFileName = GeneralUtil.fileName(oldName);
									String newName = newHomeFolder + tempFileName;
									//if the old name starts with the oldhome and the length
									//...of the old file name is the sum of the oldHome file length
									// plus the short file name plus 1 (in other words there is just a file
									//...seperator between them) then we replace that with the new home
									//...folder
									if(oldName.startsWith(oldHome) && 
											(oldHome.length() + 1 + tempFileName.length() == oldName.length() ))
									((Element)fileIndex).setAttribute(NAME_KEY, newName);
								}
								fileIndex = fileIndex.getNextSibling();
							}
						}
					groupIndex = groupIndex.getNextSibling();
					}
				}//end if
				//go to the next sibling
				opIndex = opIndex.getNextSibling();
			}//end while
		
			projectInfo.getDocumentElement().setAttribute(HOME_KEY, newHome);
			
			writeProjectInfo(true, false, false);
		}
	}
	
	private Node getOperation(String name){
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right name
				if(((Element)index).getAttribute(NAME_KEY).equals(name))
					return index;
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		return null;
	}
	
	public void swapOp(String op1, String op2){
		
		Node firstNode = getOperation(op1);
		Node secondNode = getOperation(op2);
		
		//replace the seond node with the first and store the parentless old node
		secondNode = projectInfo.getDocumentElement()
						.replaceChild(firstNode.cloneNode(true), secondNode);
		
		projectInfo.getDocumentElement()
					.replaceChild(secondNode.cloneNode(true), firstNode);
		
	}
	
	/**
	 * add a file element to a given parent parent node with their discription.
	 * @param parent node that we are attaching the file-node too
	 * @param name set the name attribute of this file-node
	 * @param discription set the text of the file node
	 */
	void addFile(Node parent, String name, String discription){
		
		String homeFolder = projectInfo.getDocumentElement().getAttribute(HOME_KEY);
		String shortname = GeneralUtil.fileName(name);
		//if file name is in the home directory then truncate
		if(name.startsWith(homeFolder) &&
				(name.length() == homeFolder.length() + shortname.length() + 1));
			name = shortname;
		
		if(parent == null
				|| parent.getNodeType() != (Node.ELEMENT_NODE)){
			new ErrorDialog(frame, "We are trying to add a file to a " +
					"node which can not accept it.");
			return;
		}
		if(name == null){
			new ErrorDialog(frame, "Error trying to add a null operation name. \n" +
					"addFile@Project.java");
			return;
		}
		
		if(discription == null){
			discription = "";
		}
		
		/*Node index = parent.getFirstChild();
		while(index != null){
			//go through and check to see if we have the file
			//...in the list already
			if(index.getNodeName().equals(FILE_KEY)
					 && name.equals((((Element)index).getAttribute(NAME_KEY)))){
				//if we do then return
				return;
			}
			index.getNextSibling();
		}
		*/
		
		//new file node
		Element newFile = (Element) projectInfo.createElement(FILE_KEY);
		//set the name attribute
		newFile.setAttribute(NAME_KEY, name);
		//set the text content
		//newFile.setTextContent(discription);
		newFile.appendChild(projectInfo.createTextNode(discription));
		//set up spacing
		parent.appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		//attach the file node
		parent.appendChild(newFile);
				
	}
	
	/**
	 * Add a file to an operation.
	 * @param opName A string that is the operation's identifying name
	 * @param fileType A string that flags weather the file is input/output
	 * @param fileName A string that gives the file name
	 * @param fileDisc A string that holds the file discription
	 * @return a boolean that is  weather the file was added or not
	 */
	public boolean addFile(String opName, String fileType, String fileName, String fileDisc){
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
//			check to see if it has the right name
			if(index.getNodeName().equals(OP_KEY)
					&& opName.equals((((Element)index).getAttribute(NAME_KEY)))){
				//look at all the nodes in the operation
				Node index2 = index.getFirstChild();
				while(index2 != null){
					//look at either the infile or outfile list
					if(index2.getNodeName().equals(fileType)){
						addFile(index2, fileName, fileDisc);
						return true;
						}
					index2 = index2.getNextSibling();
					}
				}
			
			//go to the next sibling
			index = index.getNextSibling();
		}
		return false;
		
	}
	
	/**
	 * Remove a list of files from certain operations.
	 * @param operations Vector of operation names we wish to remove files from
	 * @param filename Vector of file names we wish to remove
	 */
	public void removeFile(Vector <String> operations, Vector <String> filename){
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right name
				if(operations.contains(((Element)index).getAttribute(NAME_KEY))){
					//look at all the nodes in the operation
					Node index2 = index.getFirstChild();
					while(index2 != null){
						//look at either the infile or outfile list
						if(index2.getNodeName().equals(INFILE_KEY)
								|| index2.getNodeName().equals(OUTFILE_KEY)){
							//go through each file
							Node index3 = index2.getFirstChild();
							while(index3 != null){
								//check to see if the filename is on the nix list
								if(index3.getNodeName().equals(FILE_KEY)
										&& filename.contains(((Element)index3)
												.getAttribute(NAME_KEY))){
									//incrament the index
									Node temp = index3.getNextSibling();
									//remove the previous file entry
									index2.removeChild(index3);
									index3 = temp;
									//change the number of files here
									//int numFiles = Integer.parseInt(
									//		((Element)index2).getAttribute(NUM_FILE_KEY));
									//((Element)index2).setAttribute(NUM_FILE_KEY,
									//		(Integer.toString((numFiles -1))));
								}else{
									//just go on
									index3 = index3.getNextSibling();
								}
							}
						}
						index2 = index2.getNextSibling();
					}
				}
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
	}
	
	/**
	 * replace the discription of a file with a new discription
	 * @param parent parent node of the file we are looking for
	 * @param name name of the file we are looking for
	 * @param discription new discription of file
	 */
	void replaceFile(Node parent, String name, String discription){
		//dummy check the inputs
		if(parent == null
				|| parent.getNodeType() != (Node.ELEMENT_NODE)){
			System.err.println("We are trying to add a file to a " +
					"node which can not accept it.");
			return;
		}
		
		//go though the children of the parent node
		Node index = parent.getFirstChild();
		while(index != null){
			//check that we are at a file node with the right name
			if(index.getNodeName().equals(FILE_KEY)
					&& ((Element) index).getAttribute(NAME_KEY).equals(name)){
				//there might not be any children
				if(index.getFirstChild() != null)
					//reset the discription
					index.getFirstChild().setNodeValue(discription);
				else
					//add a new child with the discription
					index.appendChild(projectInfo.createTextNode(discription));
				return;
			}
			index = index.getNextSibling();
		}
		
		//System.err.println("Could not find the file to replace " +
		//		"the discription for.");
		return;
	}
	
	public String getFolderFileNode(String fileName){
		
//		dummy check the inputs
		if(projectInfo == null){
			System.err.println("We are trying to add a file to a " +
					"node which can not accept it.");
			return "";
		}
		
		fileName = GeneralUtil.fileName(fileName);
		
		//go though the children of the parent node
		Node index = projectInfo.getDocumentElement().getFirstChild();
		
		while(index != null){
			//go to the folder node
			if(index.getNodeName().equals(FOLDER_KEY)){
				Node index2 = index.getFirstChild();
				while(index2 != null){
//					find the fileName
					if(((Element)index2).getAttribute(SHORT_NAME_KEY)
							.equals(fileName)){
						return index2.getTextContent();
					}
					//set the discription
					//if you can't find the fileName
					//create an entry
					index2 = index2.getNextSibling();
				}
				
				return "";
			}
			index = index.getNextSibling();
		}
		
		return "";
	}
	
	public void setFolderFileNote(String fileName, String fileDisc){
//		dummy check the inputs
		if(projectInfo == null){
			System.err.println("We are trying to add a file to a " +
					"node which can not accept it.");
			return;
		}
		
		fileName = GeneralUtil.fileName(fileName);
		
		//go though the children of the parent node
		Node index = projectInfo.getDocumentElement().getFirstChild();
		
		while(index != null){
			//go to the folder node
			if(index.getNodeName().equals(FOLDER_KEY)){
				Node index2 = index.getFirstChild();
				while(index2 != null){
//					find the fileName
					if(((Element)index2).getAttribute(SHORT_NAME_KEY)
							.equals(fileName)){
						index2.replaceChild(projectInfo.createTextNode(fileDisc), 
								index2.getFirstChild());
						return;
					}
					//set the discription
					//if you can't find the fileName
					//create an entry
					index2 = index2.getNextSibling();
				}
				Element newFile = (Element) projectInfo.createElement(FILE_KEY);
				newFile.setAttribute(SHORT_NAME_KEY, fileName);
				newFile.appendChild(projectInfo.createTextNode(fileDisc));
				index.appendChild(newFile);
				return;
			}
			index = index.getNextSibling();
		}
		//we didn't find the node! so we add it
		//...new file node
		Element newFolder = (Element) projectInfo.createElement(FOLDER_KEY);
		projectInfo.getDocumentElement().appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		//attach the file node
		projectInfo.getDocumentElement().appendChild(newFolder);
		setFolderFileNote(fileName, fileDisc);
		return;
	}

	
	
	/**
	 * add an operation element to the project element.
	 * @param cline the actual command line that is sent plink
	 * @param name name that command for internal reference
	 * @param discription set the discription of that command
	 * @param input an array of names of the input files
	 * @param inputDis an array of discriptions corrisponding to the inptu files
	 * @param output an array of names of the output files
	 * @param outputDis an array of discriptions corrisponding to the output files
	 * @return return that operation element that you created incase 
	 * you want to do something with it
	 */
	public Element addOperation(String cline, String name, String discription,
						String [] input, String [] inputDis,
						String [] output, String [] outputDis){
		//create new element to hold operation
		Element newOp = (Element) projectInfo.createElement(OP_KEY);
		//set the appropreate attributes
		newOp.setAttribute(CLINE_KEY, cline);
		newOp.setAttribute(NAME_KEY, name);
		
		//set the text for the node as the discription
		newOp.setTextContent(discription);
		
		//create a new element to hold the infiles
		Element infiles = (Element) projectInfo.createElement(INFILE_KEY);
		//start with no file's
		//infiles.setAttribute(NUM_FILE_KEY, "0");
		//set a name key for the hell of it, don't know that it's actually
		//...used
		infiles.setAttribute(NAME_KEY, "Input Files");
		//go through and add each input file name and discription
		for(int i = 0; i < input.length; i ++){
			addFile(infiles, input[i], inputDis[i]);
		}
		//format so that humans can read the xml format
		newOp.appendChild( projectInfo.createTextNode(String.format("%n", (Object[]) null)));
		//add all the infiles to the new operation
		newOp.appendChild(infiles);
		
		//create the new element to hold the outfiles
		Element outfiles = (Element) projectInfo.createElement(OUTFILE_KEY);
		//outfiles.setAttribute(NUM_FILE_KEY, "0");
		outfiles.setAttribute(NAME_KEY, "Output Files");
		for(int i = 0; i < output.length; i ++){
			addFile(outfiles, output[i], outputDis[i]);
		}
		
		//format so that humans can read the xml format
		newOp.appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		newOp.appendChild(outfiles);
		//format so that humans can read the xml format
		newOp.appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		
		//get the number of operations ever created
		//int num_ever_op = Integer.parseInt(
		//	 projectInfo.getDocumentElement().getAttribute(EOP_KEY)) + 1;
		//add this to the operation count
		//projectInfo.getDocumentElement().setAttribute(EOP_KEY, "" + num_ever_op);
		
//		append the operation to the project
		projectInfo.getDocumentElement().appendChild(newOp);
		
		//writeProjectInfo(false, true);
		
		if(gplinkLog != null)
			gplinkLog.addOp(name);
		
		//return the element we have just created
		return newOp;
	}
	
	/**
	 * replace an old operation with a new one
	 * @param oldname name of the old operation
	 * @param cline new command line you want to add
	 * @param name new name of operation
	 * @param discription new discription
	 * @param input new input file names
	 * @param inputDis new input file discriptions
	 * @param output new output file names
	 * @param outputDis new output file discriptions
	 * @return return a pointer to the node that was modified
	 */
	public Node replaceOperation(String oldname, String cline, String name, 
						String discription, 
						String [] input, String [] inputDis,
						String [] output, String [] outputDis){

		//******Find the operation based on it's old name********
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//element that we are looking for
		Element renewOp = null;
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type and have the right name
			if(index.getNodeName().equals(OP_KEY)
				&& ((Element)index).getAttribute(NAME_KEY).equals(oldname)){
				renewOp = (Element)index;
				}
			//go to the next sibling
			if(index != null)
				index = index.getNextSibling();
		}
		
		if(renewOp == null){
			System.err.println("Can't find the operation to replace.");
			return null;
		}
		
		//********Reset the data***********
		//set the appropreate attributes
		renewOp.setAttribute(CLINE_KEY, cline);
		renewOp.setAttribute(NAME_KEY, name);
		
		//set the text for the node as the discription
		renewOp.getFirstChild().setNodeValue(discription);
		
		index = renewOp.getFirstChild();
		
		while(index != null){
			if(index.getNodeName().equals(INFILE_KEY)){
				for(int i = 0; i < input.length; i ++){
					replaceFile(index, input[i], inputDis[i]);
				}
			}
			if(index.getNodeName().equals(OUTFILE_KEY)){
				for(int i = 0; i < output.length; i ++){
					replaceFile(index, output[i], outputDis[i]);
				}
			}
			index = index.getNextSibling();
		}
		
		//return the element we have just replaced
		return renewOp;
		
	}
	
	/**
	 * delete an operation in the document by name
	 * @param name of the operation to be delete
	 * @return the operation that we just deleted incase we want it
	 */
	public Node removeOperation(String name){
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			//...and that the name matches
			if(index.getNodeName().equals(OP_KEY)
					&& ((Element) index).getAttribute(NAME_KEY).equals(name)){
				//found it!
				return projectInfo.getDocumentElement().removeChild(index);
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		//we didn't find the operation
		return null;
	}
	
	
	/////////////////////////////////////////////////
	/////////////// Write Data //////////////////////
	/////////////////////////////////////////////////
	
	/**
	 * This writes the data that makes up the project 
	 * (files and operations in general)
	 * to either the log file (.metafile) and/or the output screen.
	 * @param fileFlag flags the print to file output
	 * @param screenFlag flags the print to screen output
	 * 
	 */
	public void writeProjectInfo(boolean fileFlag, 
			boolean screenFlag, boolean waitFlag){
		
		if(avalibleLock() == false
				&& hasMetaLock() == false 
				&& getBrowse() == false){
			int choice = JOptionPane.showConfirmDialog(frame, 
					"Some other instance of gPLINK has the lock. " +
					"Would you like to steal the " +
					"lock to save the project information?\n " +
					"If you don't then you will be unable to add any new operations. ", 
					"Steal?", JOptionPane.YES_NO_OPTION);
			if(choice == JOptionPane.CLOSED_OPTION
					|| choice == JOptionPane.NO_OPTION){
				setBrowse(true);
				frame.menuStuff.setMenuLock(!getBrowse());
			}
		}
		
		//System.out.println("writeProjectInfo@Project.java:lock check completed... moving on.");
		
		if(getBrowse() == true){
			return;
		}
		lockMeta();
		frame.menuStuff.setMenuLock(true);
		
//		 Use a Transformer for output
		   TransformerFactory tFactory = TransformerFactory.newInstance();
		   Transformer transformer = null;
	
			//configure the transformer
			try {
				transformer = tFactory.newTransformer();
			} catch (TransformerConfigurationException e1) {
				new ErrorDialog(frame, "Error trying to create transformer.\n" +
						"writeProjectInfo@Project.java");
				return;
			}
			//set the source of the data
			DOMSource source = new DOMSource(projectInfo);
			//check the flags to see where we should output
			if(screenFlag == true){
				//set the recieving stream to the output
				StreamResult result = new StreamResult(System.out);
				try {
					transformer.transform(source, result);
				} catch (TransformerException e) {
					new ErrorDialog(frame, "Error trying to transform source to" +
							"console.\n writeProjectInfo@Project.java");
					return;
				}
			}
			if(fileFlag == true){
				//set the recieving stream to the file specified by 
				//...rootfolder/logfile
				StreamResult ofile = null;
				try {
					ofile = new StreamResult(new FileWriter(
								new File(rootfolder, LOG_FILE)) );
				} catch (IOException e) {
					new ErrorDialog(frame, "Error trying to create file stream.\n" +
						"writeProjectInfo@Project.java");
					return;
				}
				try {
					transformer.transform(source, ofile);
				} catch (TransformerException e) {
					new ErrorDialog(frame, "Error trying to transform source to" +
					"file stream.\n writeProjectInfo@Project.java");
					return;
				}
			}
			
		if(isRemote()){
			String temp1 = getRoot().getAbsolutePath() + File.separator + LOG_FILE;
			String temp2 = remotefolder;
			UploadFile uploader = new UploadFile(frame, temp1, temp2, getConn());
			uploader.start();
			if(waitFlag){
				try {
					uploader.join();
				} catch (InterruptedException e) {
					//fail quietly
				}
			}
		}
		   
	}
	
	public void writeOperations(boolean fileFlag, boolean screenFlag){
		if(fileFlag == true){

			DateFormat dateFormat = new SimpleDateFormat("yyyy-MMM-dd-HH-mm-ss");
            java.util.Date date = new java.util.Date();
            File outputFile = new File(getRoot(), OP_SUM_FILE+dateFormat.format(date)+".html");
		
			try {
		        BufferedWriter out = new BufferedWriter(new FileWriter(outputFile));
		        out.write("<html><title>" + getProjectName()+"</title><body>");
		        out.write("<h2>"+getProjectName()+"</h2>");
		        out.write("<h3>Timestamp: " + dateFormat.format(date)+ "</h3>");
		        out.write("<h3>local project file location: " + getRoot() + "</h3>");
		        if(isflaggedRemote())
		        	out.write("<h3>remote project file location: " + getRemoteFile() + "</h3>");
		        out.newLine();
		        out.write("<ul>");
		        
		        for(String opName: getAllOpName()){
		        	out.write("<li>");
		        	out.write(opName+ ": " + getOpDiscription(opName));
		        	
		        	out.newLine();
		        	out.write("<ul>");
		        	out.write("<li>"+getOpCommand(opName)+"</li>");
		        	out.newLine();
		        			        	
		        	String [][] infiles = getOpFiles(opName, INFILE_KEY);
		        	out.write("<li>INFILES<ul>");
		        	out.newLine();
		        	for(int i = 0; i < infiles[0].length; i ++){
		        		out.write("<li>"+infiles[0][i] + ": " 
		        				+ getFolderFileNode(GeneralUtil.fileName(infiles[0][i])) 
		        				+ "; "+ infiles[1][i]+"</li>");
		        		out.newLine();
		        	}
		        	out.write("</ul></li>");
		        	
		        	String [][] outfiles = getOpFiles(opName, OUTFILE_KEY);
		        	out.write("<li>OUTFILES<ul>");
		        	out.newLine();
	        		for(int i = 0; i < outfiles[0].length; i ++){
	        			if(!outfiles[0][i].endsWith(OP_STATUS_EXT)){
			        		out.write("<li>"+outfiles[0][i] + ": " 
			        				+ getFolderFileNode(GeneralUtil.fileName(outfiles[0][i])) 
			        				+ "; "+ outfiles[1][i]+"</li>");
			        		out.newLine();
	        			}
		        	}
	        		out.write("</ul></li>");
	        		
	        		out.write("</ul>");
	        		out.write("</li>");
	        		out.newLine();
		        }
		        out.write("</ul>");
        		out.newLine();
        		out.write("</body></html>");
		        out.close();
		    } catch (IOException e) {
		    	new ErrorDialog(frame, "Error trying to write the operations \n" +
		    			"summary.\n writeOperations@Project.java");
		    }
		    if(isRemote() && !getBrowse()){
		    	new UploadFile(frame, outputFile.getAbsolutePath(),
		    			getRemoteFile(), getConn()).start();
		    }
		}
		if(screenFlag == true){
			System.out.println(getProjectName());
			System.out.println();
			for(String opName: getAllOpName()){
	        	System.out.print(opName+ ": " + getOpDiscription(opName));
	        	
	        	String [][] infiles = getOpFiles(opName, INFILE_KEY);
	        	System.out.println("INFILES");
	        	for(int i = 0; i < infiles[0].length; i ++){
	        		System.out.println(infiles[0][i] + ": " + infiles[1][i]);
	        	}
	        	
	        	String [][] outfiles = getOpFiles(opName, OUTFILE_KEY);
	        	System.out.println("OUTFILES");
        		for(int i = 0; i < outfiles[0].length; i ++){
        			System.out.println(outfiles[0][i] + ": " + outfiles[1][i]);
	        	}
        		System.out.println();
	        }
		}		
	}

}
